//-< CacheManager.java >---------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      4-Jan-2001  K.A. Knizhnik  * / [] \ *
//                          Last update:  4-Jan-2001  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Object cache manager
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
	
    /// <summary> <code>CacheManager</code> may be used on a per Thread basis (see attatch 
    /// and detatch), or on a global basis (defaultCacheManager), to cache the
    /// objects on client side and to keep transaction list.
    /// Two caches are kept, a last recently used and a frequesntly used. Cache
    /// and transaction list are kept as double linked lists.
    ///
    /// </summary>
    /// <author>  K.A. Knizhnik
    /// </author>
    /// <version> 1.0
    /// 
    /// </version>
    public class CacheManager
    {
        /// <summary> <code>getCacheManager</code> the current manager. This is the per-thread
        /// manager that has called attatch() or the default one if attatch has never
        /// been called on the current thread.
        ///
        /// </summary>
        /// <returns>a <code>CacheManager</code> value, see above.
        /// 
        /// </returns>
        public static CacheManager GetCacheManager()
        {
            if (nAttached == 0)
            {
                return defaultCacheManager;
            }
            System.Threading.Thread thread = System.Threading.Thread.CurrentThread;
            CacheManager mng = (CacheManager) cacheManagerHash[thread];
            if (mng == null)
            {
                mng = defaultCacheManager;
            }
            return mng;
        }

        /// <summary> Set new watermark for minimal size of fee memory
        /// </summary>
        public virtual long FreeMemoryWatermark
        {
            set
            {
                freeMemoryWatermark = value;
            }
			
        }
        /// <summary> Critical section to synchronize access to static variables of metaobject
        /// </summary>
        internal Object globalCS;

        /// <summary> Header of l2-list of objects involved in trabsaction. Modified objects 
        /// are placed at the beginning of the list (transObjects.next...), while
        /// not-modified objects involved in transaction are placed at the end of
        /// the list (transObjects.prev...)
        /// This is done to optimize sending of transaction to the storage server
        /// (no extra sort or repacking of objects is needed).
        ///
        /// If object was first included in transaction list as "non-modified" and
        /// later was modified, then it will be relinked to the beginning of the 
        /// transaction list. 
        ///
        /// Object are kept in the transaction list until the end of transaction to 
        /// prevent GC from deallocating them before end of tranasction.
        /// </summary>
        internal ObjectHeader transObjects;

        /// <summary> Headers for two l2-lists used to implement enhanced LRU discipline
        /// for replacing objects in object cache. Head of the list contains
        /// most recently used object and tail - least recently used objects. 
        /// To prevent object cache from replacing all object as a result of
        /// database serach through large number of obejct (most of which will
        /// not be used in future), standard LRU scheme was extended to more fit
        /// specific of database applications. Object cache is divided into two 
        /// part: primary cache and cache for frequently used objects. First time 
        /// object is accessed, it is placed in primary cache. Object will be moved
        /// to "frequently used objects" cache only when it is accessed more than
        /// twice and it is not at the head of LRU list (case of several sucessive 
        /// accesses to the object). Both cache parts are managed separately: when
        /// total size of objects in the one part exceeds some limit value, least
        /// recently used objects from this part are thrown away from clients memory
        /// not affecting objects in another part of the cache.
        /// </summary>
        internal ObjectHeader lru0;

        /// <summary> Frequesntly accessed object list (see above )
        ///
        /// </summary>
        internal ObjectHeader lru1;

        /// <summary> Total sizes of objects in the part of the cache (see explanation above)
        /// Stub objects are not placed in the cache so thier size in not included
        /// into this values. The total size occpied by persistent objects in 
        /// client memory therefore can be greater than cacheSize0+cacheSize1. 
        /// </summary>
        internal int cacheSize0;

        /// <summary> size of frequently accessed list (see above)
        ///
        /// </summary>
        internal int cacheSize1;

        /// <summary> Limits for total size of objects in both parts of the cache (see
        /// explanation above). If size of some object exceed this limit values, 
        /// it will be placed in the cache in any case. 
        /// </summary>
        internal int cacheSizeLimit0 = 2 * 1024 * 1024;

        /// <summary> Limit for the frequently accessed Cache size 2Mb
        ///
        /// </summary>
        internal int cacheSizeLimit1 = 2 * 1024 * 1024;
		
        /// <summary> Minimal size of available free memory after reaching which object cache 
        /// is cleaned to prevent hanging of thrown objects in the memory as
        /// a result of presence references to this objects.  (1M)
        /// </summary>
        internal long freeMemoryWatermark = 1024 * 1024;
		
        /// <summary> Counter of nested transactions (or nested invocations of methods of
        /// classes derived from Persistent class). BasicMetaobject implements 
        /// implicit scheme of detection transaction boundaries. Transaction is 
        /// committed when no more active methods exists for the classes controlled 
        /// by this metaobject (or metaobjects derived from Basicmetaobject), 
        /// i.e. when nestedTransactionCount is equal to zero. It is possbile
        /// to manualy increment and decrement this count, so explictly
        /// set transaction boundaries (see methods beginNestedTransaction() and
        /// endNestedTransaction() below.
        /// </summary>
        internal int nestedTransactionCount;
		
        /// <summary> Head of a L1 list to save "forgotten objects". Forgotten 
        /// means they were garbage collected, they are added here in 
        /// ObjectState.finalize(). Their Id's are sent back to the server, as 
        /// finalizing is almost like locally deleting the object. This helps the 
        /// server side garbage collection.
        /// </summary>
        internal ForgottenObject forgottenObjectList;
		
        /// <summary> Mutex used to provide exclusive access to CacheManager. By means of this mutex 
        /// per-threads transactions sharing the same database connection are implemented. 
        /// This mutex is used by BasicMetaobject when transaction model is PER_THREAD
        /// </summary>
        internal Mutex transactionMutex;
		
        /// <summary> Isolation level: Database.PER_THREAD_TRANSACTION or Database.PER_PROCESS_TRANSACTION
        /// </summary>
        internal int isolationLevel;
		
        /// <summary> Initialize all the double linked lists with empty Persistent objects.
        /// The test whether a list is empty is thus list.next==list.prev .
        /// </summary>
        public CacheManager()
        {
            globalCS = new Object();
            transactionMutex = new Mutex();
            isolationLevel = Database.PER_PROCESS_TRANSACTION;
            lru0 = new ObjectHeader(null);
            lru0.next = lru0.prev = lru0;
            lru1 = new ObjectHeader(null);
            lru1.next = lru1.prev = lru1;
            transObjects = new ObjectHeader(null);
            transObjects.next = transObjects.prev = transObjects;
        }
		
        /// <summary> <code>defaultCacheManager</code> is the one used if the current thread,
        /// requesting the CacheManager (CahceManager.getCacheManager()) isnot 
        /// attatched. In other words this is used as a "global" CacheManager.
        /// </summary>
        internal static CacheManager defaultCacheManager = new CacheManager();
		
        /// <summary>  <code>nAttached</code> is the count of how many Threads have been 
        /// attatched. If this reaches the maximum (hashCleanupPeriod = 64), the
        /// attatched threads are checked (if they are still alive).
        /// </summary>
        internal static volatile int nAttached = 0;
		
        /// <summary>  <code>cacheManagerHash</code> holds the threads that have attatched
        /// themselves. A thread may use new CahceManager().attatch() to do so. 
        /// The keys are the threads, the values the managers
        /// </summary>
        internal static System.Collections.Hashtable cacheManagerHash = new System.Collections.Hashtable();
		
		
        /// <summary> After this many attatch calls (64) the attatch method checks that all
        /// attached threads (see cacheManagerHash) are still alive
        /// </summary>
        internal const int hashCleanupPeriod = 64;
		
        /// 
        /// <summary> Associate cache manager with current thread. If the cleanupPeriod has 
        /// been exceeded, check if all attatched threads are still alive.
        /// </summary>
        public virtual void  attach()
        {
            System.Threading.Thread thread = System.Threading.Thread.CurrentThread;
            if (++nAttached % hashCleanupPeriod == 0)
            {
                foreach (System.Threading.Thread thr in cacheManagerHash.Keys) 
                {
                    if (!thr.IsAlive)
                    {
                        cacheManagerHash.Remove(thr);
                    }
                }
            }
            cacheManagerHash[thread] = this;
        }
	
	internal void attach(System.Threading.Thread thread) 
        {
            cacheManagerHash[thread] = this;
        }
	
        /// <summary> Detach cache manager from current thread 
        /// </summary>
        public virtual void  detach()
        {
            detach(System.Threading.Thread.CurrentThread);
        }
		
        /// <summary> <code>detach</code> the given thread. Removes the associated cache 
        /// manager from the hashtable (cacheManagerHash)
        ///
        /// </summary>
        /// <param name="thread">a <code>Thread</code> to detatch
        /// 
        /// </param>
        public virtual void  detach(System.Threading.Thread thread)
        {
            cacheManagerHash.Remove(thread);
        }
		
        /// <summary> Link the given object after the other. Ie, open the double linked list
        /// after "after" and insert obj into it, so that after.next = obj and 
        /// the previous after.next is obj.next
        /// </summary>
        /// <param name="after">a <code>ObjectHeader</code> value, to link obj after
        /// </param>
        /// <param name="obj">a <code>ObjectHeader</code> value, to link in after "after"
        /// 
        /// </param>
        public static void  linkAfter(ObjectHeader after, ObjectHeader obj)
        {
            (obj.prev = after).next = (obj.next = after.next).prev = obj;
        }
		
        /// <summary>  <code>linkBefore</code> links "obj" before "before" such that obj.next
        /// == before.
        ///
        /// </summary>
        /// <param name="before">a <code>ObjectHeader</code> value to link object before
        /// </param>
        /// <param name="obj">a <code>ObjectHeader</code> value to link in.
        /// 
        /// </param>
        public static void  linkBefore(ObjectHeader before, ObjectHeader obj)
        {
            (obj.next = before).prev = (obj.prev = before.prev).next = obj;
        }
		
        /// <summary>  <code>unlink</code> the given object so that obj.next.prev = obj.prev
        /// and obj.prev.next = obj.next 
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> value to unlink. 
        /// 
        /// </param>
        public static void  unlink(ObjectHeader obj)
        {
            System.Diagnostics.Debug.Assert(obj.next.prev == obj && obj.prev.next == obj);
            obj.next.prev = obj.prev;
            obj.prev.next = obj.next;
            obj.next = obj.prev = null;
        }
		
        /// <summary> Size of Persitent head plus size used by finilizer and object hash 
        /// entry  (64)
        /// </summary>
        internal const int objectAllocationOverhead = 64;
		
        /// <summary> Exclude object from LRU list so making it not available for 
        /// cached objects LRU relpacing algorithm. This method is called
        /// in two situations: 
        /// 1) when object is accessed to prevent it from throwing away until
        /// it is accessed by active method.
        /// 2) when modification notification for this object was received from
        /// server and instance of deteriorated object instance is removed from
        /// clients memory.
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> value
        /// 
        /// </param>
        public virtual void  removeFromCache(ObjectHeader obj)
        {
            lock(this)
            {
                if ((obj.state & ObjectState.CACHED) != 0)
                {
                    System.Diagnostics.Debug.Assert((obj.state & (ObjectState.DESTRUCTED | ObjectState.RAW)) == 0);
                    unlink(obj);
                    if ((obj.state & ObjectState.USEFUL) == 0)
                    {
                        if ((obj.state & ObjectState.ACCESSED) != 0)
                        {
                            if (obj != lru0.next)
                            {
                                obj.state |= ObjectState.USEFUL;
                            }
                        }
                        else
                        {
                            obj.state |= ObjectState.ACCESSED;
                        }
                        cacheSize0 -= objectAllocationOverhead + obj.size();
                    }
                    else
                    {
                        cacheSize1 -= objectAllocationOverhead + obj.size();
                    }
                    obj.state &= ~ ObjectState.CACHED;
                }
            }
        }
		
        /// <summary> Remove from cache all objects from the specified database
        /// </summary>
        /// <param name="db">database which objects should be removed from cache
        /// 
        /// </param>
        public virtual void  removeFromCache(Database db)
        {
            lock(this)
            {
                ObjectHeader obj, next;
                for (obj = transObjects.next; obj != transObjects; obj = next)
                {
                    next = obj.next;
                    if (obj.storage.database == db)
                    {
                        unlink(obj);
                        obj.opid = 0;
						
                        if (obj.data != null)
                        {
                            cacheSize0 -= objectAllocationOverhead + obj.data.Length + obj.desc.fixedSize;
                            obj.data = null;
                        }
                    }
                }
                for (obj = lru0.next; obj != lru0; obj = next)
                {
                    next = obj.next;
                    if (obj.storage.database == db)
                    {
                        unlink(obj);
                        cacheSize0 -= objectAllocationOverhead + obj.size();
                        obj.opid = 0;
                        if (obj.data != null)
                        {
                            cacheSize0 -= objectAllocationOverhead + obj.data.Length + obj.desc.fixedSize;
                            obj.data = null;
                        }
                    }
                }
                for (obj = lru1.next; obj != lru1; obj = next)
                {
                    next = obj.next;
                    if (obj.storage.database == db)
                    {
                        unlink(obj);
                    }
                }
            }
        }
		
		
        /// <summary> <code>insertInCache</code> insert object in one of two cache LRU lists
        /// depending on its current
        /// state. If total size of objects in this part of the cache exceeds limit 
        /// value, least recently used objects will be removed from the cache
        /// (and from client memory) until total size of objects becomes less or
        /// equal than limit.
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> value
        /// 
        /// </param>
        public virtual void  insertInCache(ObjectHeader obj)
        {
            lock(this)
            {
                System.Diagnostics.Debug.Assert((obj.state & (ObjectState.CACHED | ObjectState.DESTRUCTED | ObjectState.RAW)) == 0);
                //long freeMemory = GC.GetTotalMemory(false);
                if ((obj.state & ObjectState.USEFUL) == 0)
                {
                    cacheSize0 += objectAllocationOverhead + obj.size();
                    while ((/*freeMemory < freeMemoryWatermark ||*/ cacheSize0 > cacheSizeLimit0) && lru0.prev != lru0)
                    {
                        ObjectHeader victim = lru0.prev;
                        cacheSize0 -= objectAllocationOverhead + victim.size();
                        victim.metaobject.destroyObject(victim);
                        victim.state &= ~ ObjectState.CACHED;
                        unlink(victim);
                    }
                    linkAfter(lru0, obj);
                }
                else
                {
                    cacheSize1 += objectAllocationOverhead + obj.size();
                    while ((/*freeMemory < freeMemoryWatermark ||*/ cacheSize1 > cacheSizeLimit1) && lru1.prev != lru1)
                    {
                        ObjectHeader victim = lru1.prev;
                        cacheSize1 -= objectAllocationOverhead + victim.size();
                        victim.metaobject.destroyObject(victim);
                        victim.state &= ~ ObjectState.CACHED;
                        unlink(victim);
                    }
                    linkAfter(lru1, obj);
                }
                obj.state |= ObjectState.CACHED;
            }
        }
		
        /// <summary> Include object in transaction list. If object is modified it will be
        /// placed at the head of the list, otherwise at the tail of the list.
        /// If object already was included in transaction as "read-only", and 
        /// later was modified, it will relinked to the head of the list.
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> value
        /// 
        /// </param>
        public virtual void  addToTransaction(ObjectHeader obj)
        {
            lock(transObjects)
            {
                if ((obj.state & ObjectState.DIRTY) != 0)
                {
                    if ((obj.state & ObjectState.TRANSWRITE) == 0)
                    {
                        if ((obj.state & ObjectState.TRANSREAD) != 0)
                        {
                            obj.state &= ~ ObjectState.TRANSREAD;
                            unlink(obj);
                        }
                        linkAfter(transObjects, obj);
                        obj.state |= ObjectState.TRANSWRITE;
                    }
                }
                else
                {
                    if ((obj.state & ObjectState.TRANSREAD) == 0)
                    {
                        linkBefore(transObjects, obj);
                        obj.state |= ObjectState.TRANSREAD;
                    }
                }
            }
        }
		
        /// <summary> Set new limits for object cache. See description of cache management
        /// algorithm at the beggining of this file.
        /// </summary>
        /// <param name="l0">an <code>int</code> value
        /// </param>
        /// <param name="l1">an <code>int</code> value
        /// 
        /// </param>
        public virtual void  setCacheLimits(int l0, int l1)
        {
            lock(globalCS)
            {
                cacheSizeLimit0 = l0;
                cacheSizeLimit1 = l1;
            }
        }
		
        /// <summary> This method is called by Metaobject.preDaemon to determine start of access to the database
        /// </summary>
        public void  enter()
        {
            if (isolationLevel == Database.PER_THREAD_TRANSACTION)
            {
                transactionMutex.enter();
            }
        }
		
        /// <summary> This method is called by Metaobject.postDaemon to determine end of access to the database
        /// </summary>
        public void  leave()
        {
            if (isolationLevel == Database.PER_THREAD_TRANSACTION)
            {
                transactionMutex.leave();
            }
        }
		
        /// <summary> Place object in forgotten list
        /// </summary>
        public virtual void  forgetObject(ObjectHeader obj)
        {
            lock(this)
            {
                if (obj.data != null)
                {
                    cacheSize0 -= objectAllocationOverhead + obj.data.Length + obj.desc.fixedSize;
                    obj.data = null;
                }
                forgottenObjectList = new ForgottenObject(obj, forgottenObjectList);
            }
        }
    }
}