//-< WeakHashTable.java >--------------------------------------------*--------*
// GOODS                      Version 2.0        (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:     11-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update: 11-Oct-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Hash table based on weak pointers
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
	
    public class WeakHashTable : ObjectCache
    {
        internal Entry[] table;
        internal const float loadFactor = 0.75f;
        internal int count;
        internal int threshold;
		
        public WeakHashTable(int initialCapacity)
        {
            threshold = (int) (initialCapacity * loadFactor);
            if (initialCapacity != 0)
            {
                table = new Entry[initialCapacity];
            }
        }
		
        public WeakHashTable():this(10001)
        {
        }
		
        public virtual bool remove(int oid)
        {
            lock(this)
            {
                Entry[] tab = table;
                int index = (oid & 0x7FFFFFFF) % tab.Length;
                for (Entry e = tab[index], prev = null; e != null; prev = e, e = e.next)
                {
                    if (e.oid == oid)
                    {
                        count -= 1;
                        if (prev != null)
                        {
                            prev.next = e.next;
                        }
                        else
                        {
                            tab[index] = e.next;
                        }
                        return true;
                    }
                }
                return false;
            }
        }
		
        public virtual void  put(int oid, Object obj)
        {
            lock(this)
            {
                Entry[] tab = table;
                int index = (oid & 0x7FFFFFFF) % tab.Length;
                for (Entry e = tab[index]; e != null; e = e.next)
                {
                    if (e.oid == oid)
                    {
                        e.oref.Target = obj;
                        return ;
                    }
                }
                if (count >= threshold)
                {
                    // Rehash the table if the threshold is exceeded
                    rehash();
                    tab = table;
                    index = (oid & 0x7FFFFFFF) % tab.Length;
                }
				
                // Creates the new entry.
                tab[index] = new Entry(oid, new WeakReference(obj), tab[index]);
                count++;
            }
        }
		
        public virtual Object get(int oid)
        {
            lock(this)
            {
                Entry[] tab = table;
                int index = (oid & 0x7FFFFFFF) % tab.Length;
                for (Entry e = tab[index]; e != null; e = e.next)
                {
                    if (e.oid == oid)
                    {
                        if (e.oref.IsAlive) 
                        {
                            return e.oref.Target;    
                        } 
                        else 
                        {
                            remove(oid);
                            return null;
                        }    
                    }
                }
                return null;
            }
        }
		
        internal void rehash()
        {
            int oldCapacity = table.Length;
            Entry[] oldMap = table;
            int i;
            int released = 0;
            for (i = oldCapacity; --i >= 0; )
            {
                for (Entry prev = null, e = oldMap[i]; e != null; e = e.next)
                {
                    if (!e.oref.IsAlive)
                    {
                        count -= 1;
                        released += 1;
                        if (prev == null)
                        {
                            oldMap[i] = e.next;
                        }
                        else
                        {
                            prev.next = e.next;
                        }
                    }
                    else
                    {
                        prev = e;
                    }
                }
            }
            if ((uint)count <= ((uint)threshold >> 1))
            {
                return ;
            }
            int newCapacity = oldCapacity * 2 + 1;
            Entry[] newMap = new Entry[newCapacity];
			
            threshold = (int) (newCapacity * loadFactor);
            table = newMap;
			
            for (i = oldCapacity; --i >= 0; )
            {
                for (Entry old = oldMap[i]; old != null; )
                {
                    Entry e = old;
                    old = old.next;
					
                    int index = (e.oid & 0x7FFFFFFF) % newCapacity;
                    e.next = newMap[index];
                    newMap[index] = e;
                }
            }
        }
		
        /// <summary> Clears this hashtable so that it contains no keys. 
        /// </summary>
        public virtual void  clear()
        {
            lock(this)
            {
                Entry[] tab = table;
                for (int index = tab.Length; --index >= 0; )
                {
                    tab[index] = null;
                }
                count = 0;
            }
        }
		
        public virtual int size()
        {
            return count;
        }
		
        internal class Entry
        {
            internal Entry next;
            internal WeakReference oref;
            internal int oid;
			
            internal Entry(int oid, WeakReference oref, Entry chain)
            {
                next = chain;
                this.oid = oid;
                this.oref = oref;
            }
        }
    }
}