//-< Metaobject.java >-----------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      1-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update:  3-Oct-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Abstraction of metaobject protocol used by JAVAMOP preprocessor
//-------------------------------------------------------------------*--------*

package goodsjpi;

/**
 * <code>Metaobject</code> is a base class for all Transaction + locking strategies.
 * Each @see Persistent holds a MetaObject, which is called when the object is 
 * modified or read. This also means you can have different strategies for different
 * classes, or even object if you wish. The default is PessimisticRepeatableRead <br>
 * This class has abstract functions to handle loading and
 * unloading of Object over the wire, and to handle Transactions. Specifically
 * their failure strategies are implemented by different implementations of the 
 * various derived classes. <br>
 * Most functions are implemented in the BasicMetaObject, but the begin/end * Access
 * functions are implemented in the different derived classes. <br>
 * JAVAMOP, the MetaObjectProtocol postprocessor inserts all necessary code
 * to call MetaObject functions into the compiled byte code.
 *
 * @author K.A. Knizhnik
 * @version 1.0
 */
public abstract class Metaobject implements java.io.Serializable { 
  /**
   * Invocation of this <code>preDaemon</code> is inserted by MOP generator 
   * before each method invocation or object component access. (Other database
   * implementation call this method fetch(), to indicate that this causes
   * the data of the Object to be fetched, if it's not already there) <br>
   * The exact call that is inserted is  super.metaobject.preDaemon(this, attr);
   *
   * @param obj The <code>Object</code> that is to be loaded. Really it is an instance
   * of Persistent
   * @param attr an <code>int</code> value, indicating the kind of access. An OR 
   * combination of the static flags below.
   */
    abstract public void preDaemon(Object obj, int attr);

  /**
   * Invocation of this <code>postDaemon</code> is inserted by MOP generator after 
   * each  method invocation or object component access.<br>
   * The inserted code is something like:  super.metaobject.postDaemon(this,attr,flag)
   *
   * @param obj an <code>Object</code>, or really a Persistent, that has been changed
   * @param attr an <code>int</code>, indicating where/how the change happened. An OR
   * combination of the static flags below
   * @param modified a <code>boolean</code> value, indicating whether there has 
   * actually been a change. Because of exceptions it isn't always true.
   */
    abstract public void postDaemon(Object obj, int attr, boolean modified);

  /**
   *  <code>MUTATOR</code> == 1 , means object can be changed
   */
    final public static int MUTATOR    = 1; 

  /**
   *  <code>VARIABLE</code> == 2 , means access to non-self instance variable 
   */
    final public static int VARIABLE   = 2; 

  /**
   *  <code>CONSTRUCTOR</code> == 4 means  wrapped method is constructor
   */
    final public static int CONSTRUCTOR= 4; 

  /**
   *  <code>EXCEPTION</code> == 8, means method is terminated by exception 
   */
    final public static int EXCEPTION  = 8; 

  /**
   *
   * <code>mutator</code> is only hint to MOP preprocessor to consider method
   * invoking Metaobject.mutator() as been mutator. No actual code
   * will appear in preprocessed class file.
   */
    public static void mutator() {} 

  /**
   * <code>modify</code>  provides MOP preprocessor with information that objects
   *  was (or will be) modified. No method invocation will appear in 
   * preprocessed code.
   */
    public static void modify() {} 

  /**
   *  <code>preloadObject</code> is called by Storage.read()  when it's bytes
   * have been received and just before they are copied into it's data member <br>
   *
   * @param obj a <code>Persistent</code> that has been read (from the net)
   * @param size an <code>int</code>, how many bytes the data member will hold
   */
    abstract protected void preloadObject(Persistent obj, int size);

  /**
   *  <code>loadObject</code> is not so much the loading of the Object, ,but a 
   * notification thereof. The object is really loaded by the Storage class  
   * (in prepare(obj) method) and after the deed is done, this method is called. 
   *
   * @param obj a <code>Persistent</code> object that has been loaded
   */
    abstract protected void loadObject(Persistent obj);

  /**
   * <code>forgetObject</code> is called when the objects finalize() is called, if 
   * it's opid is still valid. If all data were local, this would mean the object
   * could be destroyed because it's not reachable anymore. But because more 
   * references may exist at the servers end, this basically reduces the reference 
   * count at the servers side. <br>
   * Server side garbage collection will eventually reclaim the object space, if
   * it really has become unreachable. <br>
   *
   * @param obj a <code>Persistent</code> that has become locally unreachable
   */
    abstract protected void forgetObject(Persistent obj);

  /**
   * <code>releaseObject</code> is called during a commitTransaction, for objects
   * that have not been modified.
   *
   * @param obj a <code>Persistent</code> to be released.
   */
    abstract protected void releaseObject(Persistent obj);

  /**
   *  <code>commitObjectChanges</code> is called during commitTransaction. 
   *
   * @param obj a <code>Persistent</code> object to commit
   */
    abstract protected void commitObjectChanges(Persistent obj);

  /**
   * <code>undoObjectChanges</code> is called during an abortTransaction. <br>
   *
   * @param obj a <code>Persistent</code> value
   */
    abstract protected void undoObjectChanges(Persistent obj);

  /**
   * <code>makePersistent</code> should do just that, make the obj persistent.
   *
   * @param obj a <code>Persistent</code> object to make persistent (or prepare for
   *   that)
   * @param parent a <code>Persistent</code> object "close" to the obj. May be 
   *  used , as by BasicMetaObject, to place the obj in the same storage as parent.
   */
    abstract protected void makePersistent(Persistent obj, Persistent parent);

  /**
   *  <code>beginReadAccess</code> is called when the read Access begins. It is called
   * from the (auto generated) preDeamon (in BasicMataObject) if the Mutator flag is 
   * set. <br>
   * Different implementations of this function can be found in the subclasses and 
   * by eg locking or not locking the Object provide different strategies.
   *
   * @param obj a <code>Persistent</code> object that is about to be read.
   */
    abstract protected void beginReadAccess(Persistent obj);

  /**
   * <code>beginWriteAccess</code> is called when the write Access begins. It is 
   * called from the (autogenerated) preDeamon (in BasicMataObject) if the Mutator
   *  flag is not set. <br>
   * Different implementations of this function can be found in the subclasses and 
   * by eg locking or not locking the Object provide different strategies.
   *
   * @param obj a <code>Persistent</code> object that is about to be read.
   */
    abstract protected void beginWriteAccess(Persistent obj);
    
  /**
   *  <code>endAccess</code> is called when the (current) access/function has ended, 
   * and the object is not accessed anymore (the accesscount hits 0 ).
   *  It is called from the (autogenerated) postDeamon in BasiMetaObject.
   *
   * Different implementations of this function can be found in the subclasses and 
   * by eg locking or not locking the Object provide different strategies.
   *
   * @param obj a <code>Persistent</code> object that is about to be read.
   */
    abstract protected void endAccess(Persistent obj);

  /**
   * This method is asynchronously called by server storag agent when
   * deterioration notification from server is received. <br>
   *
   * @param obj a <code>Persistent</code> object to be invalidated.
   */
    abstract protected void invalidateObject(Persistent obj);    

  /**
   * Destoroy all object references to make it possible to provide
   * more unaccessible objects to GC <br>
   *
   * @param obj a <code>Persistent</code> value
   */
    abstract protected void destroyObject(Persistent obj);

  /**
   * Enable or disable notification on receving invalidation messages from
   * the server for this object. 
   *
   * @param obj a <code>Persistent</code> to make the notification changes for
   * @param status a <code>boolean</code>, whether notifications are used for this
   * object
   */
    abstract protected void notifyOnModification(Persistent obj, boolean status);
}

    
