//-< RtreePage.java >------------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// GOODS Persistent Class Library                                    *  /  \  *
//                          Created:     17-Mar-2005  K.A. Knizhnik  * / [] \ *
//                          Last update: 17-Mar-2005  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// R-Tree page
//-------------------------------------------------------------------*--------*

package goodslib;

import java.util.ArrayList;

class RtreePage extends SpatialObject { 
    static final int card = (4096 - 8*4 - 4)/6;
    static final int minFill = card/2;

    int             n;
    SpatialObject[] branch;


    RtreePage(SpatialObject obj) {
        branch = new SpatialObject[card];
        branch[0] = obj;
        setRectangle(obj.getRectangle());
        n = 1;
    }
    
    RtreePage(RtreePage root, RtreePage p) { 
        branch = new SpatialObject[card];
        n = 2;
        branch[0] = root;
        branch[1] = p;
        setRectangle(root.getRectangle().join(p.getRectangle()));
    }

    RtreePage() {}

    Rectangle getRectangle(int i) { 
        return branch[i].getRectangle();
    }

    RtreePage insert(SpatialObject obj, int level) {
        if (--level != 0) { 
            Rectangle r = obj.getRectangle();
            // not leaf page
            int i, mini = 0;
            double minIncr = Double.MAX_VALUE;
            double minArea = Double.MAX_VALUE;
            for (i = 0; i < n; i++) { 
                Rectangle ri = getRectangle(i);
                double area = ri.area();
                double incr = Rectangle.joinArea(ri, r) - area;
                if (incr < minIncr) { 
                    minIncr = incr;
                    minArea = area;
                    mini = i;
                } else if (incr == minIncr && area < minArea) { 
                    minArea = area;
                    mini = i;
                }                    
            }
            RtreePage p = (RtreePage)branch[mini];
            RtreePage q = p.insert(obj, level);
            setRectangle(getRectangle().join(r));
            if (q == null) { 
                // child was not split
                return null;
            } else { 
                // child was split
                return addBranch(q);
            }
        } else { 
            return addBranch(obj);
        }
    }

    int remove(SpatialObject obj, int level, ArrayList reinsertList) {
        if (--level != 0) { 
            Rectangle r = obj.getRectangle();
            for (int i = 0; i < n; i++) { 
                if (r.intersects(getRectangle(i))) { 
                    RtreePage pg = (RtreePage)branch[i];
                    int reinsertLevel = pg.remove(obj, level, reinsertList);
                    if (reinsertLevel >= 0) { 
                        if (pg.n >= minFill) { 
                            setRectangle(cover());
                        } else {
                            // not enough entries in child
                            reinsertList.add(pg);
                            reinsertLevel = level - 1;
                            removeBranch(i);
                        }
                        return reinsertLevel;
                    }
                }
            }
        } else {
            for (int i = 0; i < n; i++) { 
                if (branch[i] == obj) { 
                    removeBranch(i);
                    return 0;
                }
            }
        }
        return -1;        
    }


    void find(Rectangle r, ArrayList result, int level) {
        if (--level != 0) { /* this is an internal node in the tree */
            for (int i = 0; i < n; i++) { 
                if (r.intersects(getRectangle(i))) {
                    ((RtreePage)branch[i]).find(r, result, level); 
                }
            }
        } else { /* this is a leaf node */
            for (int i = 0; i < n; i++) { 
                if (r.intersects(getRectangle(i))) { 
                    result.add(branch[i]);
                }
            }
        }
    }

    void purge(int level) {
        if (--level != 0) { /* this is an internal node in the tree */
            for (int i = 0; i < n; i++) { 
                ((RtreePage)branch[i]).purge(level);
            }
        }
    }

    final void removeBranch(int i) {
        n -= 1;
        System.arraycopy(branch, i+1, branch, i, n-i);
        branch[n] = null;
        setRectangle(cover());
    }

    final RtreePage addBranch(SpatialObject obj) { 
        if (n < card) { 
            branch[n++] = obj;
            setRectangle(getRectangle().join(obj.getRectangle()));
            return null;
        } else { 
            return splitPage(obj);
        }
    }

    final RtreePage splitPage(SpatialObject obj) { 
        int i, j, seed0 = 0, seed1 = 0;
        double[] rectArea = new double[card+1];
        double   waste;
        double   worstWaste = Double.MIN_VALUE;
        //
        // As the seeds for the two groups, find two rectangles which waste 
        // the most area if covered by a single rectangle.
        //
        Rectangle r = obj.getRectangle();
        rectArea[0] = r.area();
        for (i = 0; i < card; i++) { 
            rectArea[i+1] = getRectangle(i).area();
        }
        Rectangle bp = r;
        for (i = 0; i < card; i++) { 
            for (j = i+1; j <= card; j++) { 
                waste = Rectangle.joinArea(bp, getRectangle(j-1)) - rectArea[i] - rectArea[j];
                if (waste > worstWaste) {
                    worstWaste = waste;
                    seed0 = i;
                    seed1 = j;
                }
            }
            bp = getRectangle(i);
        }       
        byte[] taken = new byte[card];
        Rectangle group0, group1;
        double    groupArea0, groupArea1;
        int       groupCard0, groupCard1;
        RtreePage pg;

        taken[seed1-1] = 2;
        group1 = getRectangle(seed1-1);

        if (seed0 == 0) { 
            group0 = r;
            pg = new RtreePage(obj);
        } else { 
            group0 = getRectangle(seed0-1);
            pg = new RtreePage(branch[seed0-1]);
            branch[seed0-1] = obj;
        }
        groupCard0 = groupCard1 = 1;
        groupArea0 = rectArea[seed0];
        groupArea1 = rectArea[seed1];
        //
        // Split remaining rectangles between two groups.
        // The one chosen is the one with the greatest difference in area 
        // expansion depending on which group - the rect most strongly 
        // attracted to one group and repelled from the other.
        //
        while (groupCard0 + groupCard1 < card + 1 
               && groupCard0 < card + 1 - minFill
               && groupCard1 < card + 1 - minFill)
        {
            int betterGroup = -1, chosen = -1;
            double biggestDiff = -1;
            for (i = 0; i < card; i++) { 
                if (taken[i] == 0) { 
                    Rectangle ri = getRectangle(i);
                    double diff = (Rectangle.joinArea(group0, ri) - groupArea0)
                              - (Rectangle.joinArea(group1, ri) - groupArea1);
                    if (diff > biggestDiff || -diff > biggestDiff) { 
                        chosen = i;
                        if (diff < 0) { 
                            betterGroup = 0;
                            biggestDiff = -diff;
                        } else { 
                            betterGroup = 1;
                            biggestDiff = diff;
                        }
                    }
                }
            }
            Rectangle rb = getRectangle(chosen);
            if (betterGroup == 0) { 
                group0.join(rb);
                groupArea0 = group0.area();
                taken[chosen] = 1;
                pg.branch[groupCard0++] = branch[chosen];
            } else {
                groupCard1 += 1;
                group1.join(rb);
                groupArea1 = group1.area();
                taken[chosen] = 2;
            }
        }
        //
        // If one group gets too full, then remaining rectangle are
        // split between two groups in such way to balance cards of two groups.
        //
        if (groupCard0 + groupCard1 < card + 1) { 
            for (i = 0; i < card; i++) { 
                if (taken[i] == 0) { 
                    if (groupCard0 >= groupCard1) { 
                        taken[i] = 2;
                        groupCard1 += 1;
                    } else { 
                        taken[i] = 1;
                        pg.branch[groupCard0++] = branch[i];               
                    }
                }
            }
        }
        pg.n = groupCard0;
        n = groupCard1;
        for (i = 0, j = 0; i < groupCard1; j++) { 
            if (taken[j] == 2) {
                branch[i++] = branch[j];
            }
        }
        pg.setRectangle(pg.cover());
        setRectangle(cover());
        return pg;
    }   

    final Rectangle cover() {
        Rectangle r = getRectangle(0);
        for (int i = 1; i < n; i++) { 
            r.join(getRectangle(i));
        }
        return r;
    }
}
