package org.dcg.datamgr;

import java.util.Stack;

import org.dcg.datamgr.databasejobs.DatabaseJob;
import org.dcg.datamgr.databasejobs.DatabaseReadJob;
import org.dcg.datamgr.databasejobs.DatabaseWriteJob;
import org.dcg.gen.ThreadPoolMgr;
import org.dcg.gen.ThreadPoolMgrStats;
import org.dcg.gen.ThreadPoolWorker;

/**
 * The DatabaseThreadPoolMgr is responsible for managing the read and write threads to a database.</ br>
 * It is inherited from the ThreadPoolMgr which is responsible for the implementation of the read-threads.</ br>
 * The internal WritePoolThread is basically the implementation of the write-thread.</ br>
 * The read and write threads are seperated for usability reasons.</ br>
 * It is not always as advantageous to have multiple write threads (single processor machine) whilst this is the case forread threads.</ br>
 * 
 * @author dacosta
 */
public class DatabaseThreadPoolMgr extends ThreadPoolMgr {
    protected Stack _pendingWriteJobs = new Stack(); // A special queue of pending WRITE jobs for a database
    protected Stack _availableWriteThreads = new Stack(); // A queue of all available (WRITE to database) threads
    protected ThreadGroup _totalWriteThreadGroup = new ThreadGroup("writeTG"); // ThreadGroup to control the writeThreads within the DatabasePoolMgr

    protected int _maxWriteThreads = -1; // max number of threads that can be created for database WRITING purposes (best to keep this the same as the number of processors available)
    protected int _minWriteThreads = -1; // min number of threads that must always be present for database WRITING purposes
    protected int _nrOfWriteThreadInstances = 0; // Counter to keep track of the instances created

    /**
     * Each thread in the pool is an instance of this class.<br>
     * This thread-instance is actually 'the thing' that executes the job handed to it.<br>
     * An example would be an employee in a company.<br>
     * Basically this is the 'worker/ labour class' that is available to the Mgr/ the pool
     */
    class WritePoolThread extends ThreadPoolWorker {
        
        /**
         * Constructor for PoolThread as 'normal' PoolMgr.<br>
         * @param tg is the ThreadGroup to which this thread belongs
         * @param lock is the object that takes care of overall PoolMgr locking synchronisation
         */
        public WritePoolThread(ThreadGroup tg, Object lock, boolean daemonIndicator) {
            super(tg, lock, daemonIndicator);
            setParentAttributes();
            this.start();
        }

        /**
         * Constructor for PoolThread as DBConnectionMgr.<br>
         * 
         * @param tg is the ThreadGroup to which this thread belongs.<br>
         * @param lock is the object that takes care of overall PoolMgr locking/ synchronisation.<br>
         * @param dbConnector contains the database info required to set the PoolMgr up as a DBConnectionMgr.<br>
         */
        public WritePoolThread(ThreadGroup tg, Object lock, boolean daemonIndicator, DBConnector dbConnector) {
            super(tg, lock, daemonIndicator, dbConnector);
            setParentAttributes();
            if (__debug) System.out.println(this.toString()+" constructor about to start");
            this.start();
        }

        protected void setParentAttributes() { // Double check
            _nrOfThreadInstances++;
            __threadCounter = _nrOfThreadInstances;
            __debug = _debug;
            __maxIdleTime = _maxIdleTime;
            __minThreads = _minThreads;
            __maxThreads = _maxThreads;
        }

        protected boolean currentThreadInAvailableThreads(ThreadPoolWorker pw) {
            return _availableWriteThreads.contains(this);
        }

        protected int getAvailableThreads() {
            return _availableWriteThreads.size();
        }

        protected int getNrOfPendingJobs() {
            return _pendingWriteJobs.size();
        }

        protected Object getPendingJob() {
                return (Object) _pendingWriteJobs.pop();
        }

        protected void pushIntoAvailableThreads(ThreadPoolWorker pw) {
            _availableWriteThreads.push(this);
        }

        protected void removeCurrentThreadFromQueue(ThreadPoolWorker pw) {
            removeMe(this);
        }

    };


	public DatabaseThreadPoolMgr(java.util.Properties props) throws NumberFormatException, IllegalArgumentException
	{
		super(props);
        Object o;
        int n;

        // Set maxWriteThreads
        o = props.getProperty("maxWriteThreads");
        if (o != null) {
            n = java.lang.Integer.parseInt((String) o);
            if (n <= 0) throw new IllegalArgumentException(this.getClass().getName()+".Constructor()=> maxThreads must be an integral value greater than 0");
            _maxWriteThreads = n;
        }

        // Set minWriteThreads
        o = props.getProperty("minWriteThreads");
        if (o != null) {
            n = java.lang.Integer.parseInt((String) o);
            // make sure the minThreads has a proper value
            if (n < 0) {
                n = 0;
            } else
                if (n > _maxWriteThreads)
                    n = _maxWriteThreads;
            // Now that minThreads has been determined the threads can be created assuming minThreads > 0
            _minWriteThreads = n;
            // create the min number of threads in advance of the system actually starting up
            if (_minWriteThreads > 0) {
                // If a thread belongs to the minimal set, it should also be a daemon.
                boolean daemonThread = true;
                for (int i = 1; i <= _minWriteThreads; i++) {
                    if (_debug) System.out.println("Creating minimum Thread nr " + (_nrOfWriteThreadInstances));
                    //PoolThread pt = new PoolThread(_totalThreadGroup, this);
                    createNewWriteThread(_totalWriteThreadGroup, daemonThread, _databaseConnector);
                    if (_debug) System.out.println("Size of the jobQueue "+ _pendingWriteJobs.size()+ " [req 0]");
                }
                if (_debug) System.out.println("Size of the queue "+ _availableWriteThreads.size()+ " [req "+ _minWriteThreads+ "]");
            }
        }
    }


	/**
	 * @see org.misc.util.ThreadPoolMgr#addJob(DatabaseJob)
	 */
	synchronized public void addJob(DatabaseJob job) {
        // Start by adding the job to be executed to the proper pending list
        if (job instanceof DatabaseReadJob) {
            _pendingJobs.push(job);
            // Check to see whether there is a thread available to give the job to
            if (!_availableThreads.isEmpty()) { // there are threads available
                if (_debug) System.out.println("Using an existing thread...");
                // Get a thread from the stack to give the work to
                PoolThread pt;
                synchronized (pt = (PoolThread) _availableThreads.pop()) {
                    if (_debug) System.out.println(pt.getName() + "lowered threads in the queue to:"+ _availableThreads.size());
                    // wake the thread so that it can start reading from the pending jobs list
                    pt.wake(new String("Wake up call for " + pt.getName()));
                }
                return;
            } else { // NO threads are readily available so check to see whether more can be added
                if (_maxThreads == -1 || _totalThreadGroup.activeCount() < _maxThreads) {
                    // infinite number of threads are allowed or new additional threads can be added
                    if (_debug) System.out.println("Creating additional Thread nr " + (_nrOfThreadInstances));
                    // As soon as the thread gets constructed he starts working on the pending queue
                    // False indicates it is NOT a daemon thread
                    createNewThread(_totalThreadGroup, false, _databaseConnector);
                    return;
                } else { // No threads available and no more can be created
                    // When one of the busy threads is done, it will check the pending queue and will see to this job.
                    if (_debug) System.out.println("Job has been placed in the queue because all threads in the pool are busy.");
                    return;
                }
            }
        } else if (job instanceof DatabaseWriteJob) {
            _pendingWriteJobs.push(job);
            if (!_availableWriteThreads.isEmpty()) { // Threads are available for use
                if (_debug) System.out.println("Using an existing thread...");
                WritePoolThread wpt;
                synchronized (wpt = (WritePoolThread) _availableWriteThreads.pop()) {
                    if (_debug) System.out.println(wpt.getName() + "lowered threads in the queue to:"+ _availableWriteThreads.size());
                    wpt.wake(new String("Wake up call for " + wpt.getName()));
                }
                return;
            } else {
                if (_maxWriteThreads == -1 || _totalWriteThreadGroup.activeCount() < _maxWriteThreads) {
                    if (_debug) System.out.println("Creating additional Thread nr " + (_nrOfWriteThreadInstances));
                    createNewWriteThread(_totalWriteThreadGroup, false, _databaseConnector);
                    return;
                } else {
                    if (_debug) System.out.println("Job has been placed in the queue because all threads in the pool are busy.");
                    return;
                }
            }
        }
	} // synchronized public void addJob(DatabaseJob job)

    
	synchronized public ThreadPoolMgrStats getStats() {
		super.getStats();
		//stats.maxThreads = _maxThreads;
		stats.maxWriteThreads = _maxWriteThreads;
		//stats.minThreads = _minThreads;
		stats.minWriteThreads = _minWriteThreads;
		//stats.maxIdleTime = _maxIdleTime;
		//stats.pendingJobs = _pendingJobs.size();
		stats.pendingWriteJobs = _pendingWriteJobs.size();
		//stats.numThreads = _totalThreadGroup.activeCount();
		stats.numWriteThreads = _totalWriteThreadGroup.activeCount();
		//stats.jobsInProgress = 0;
		stats.writeJobsInProgress = 0;
		//stats.databaseConnector = "empty";

		// Stats for the write-threads
		int len = this._totalWriteThreadGroup.activeCount();
		Thread[] tarray = new Thread[len];
		int __stop = _totalWriteThreadGroup.enumerate(tarray);
		for (int i = 0; i < len-1; i++) { // len-1 because the array starts counting on 0 NOT 1
			if ((tarray[i].getName().startsWith("Thread-")) && (((ThreadPoolWorker) tarray[i])._idle == false)) {
				System.out.println("\tgetStats()=>Processing thread: " + tarray[i].getName());
				stats.writeJobsInProgress++;
			}
		}

		return (stats);
	}


    /**
     * This method is called when a new WRITE thread needs to be created
     * @param tg is the ThreadGroup of which the Thread is to be a part of, null indicates NO database thread
     */
    private WritePoolThread createNewWriteThread(final ThreadGroup tg, final boolean daemonThread, final DBConnector dbConnector) {
        if (dbConnector==null) {
            // Log the fact that an empty DBConnector exists
            return null;
        } else {
            return new WritePoolThread(tg, this, daemonThread, dbConnector);
        }
    }


}
